/***************************************************************************/
/* 		This code is part of Nscache - viewer of Netscape(tm)	   */
/*		browsers disk cache					   */
/*		Copyright (c) 1999,2000 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		created 2009 by Harald Foerster				   */
/*		(harald_foerster@users.sourceforge.net)			   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include "xsearch.h"

#ifdef HAVE_XSEARCH

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "gui-gtk.h"
#include "nscache.h"

/*
#include <glib.h>
#include "charconv.h"
#include "nls.h"
#include "unicode-tools.h"
*/


#ifndef PRE_INCREMENT
#ifdef __powerpc__
#define PRE_INCREMENT		1
#endif
#endif

#if G_BYTE_ORDER == G_LITTLE_ENDIAN
static const gchar utf32_byteorder_str[] = "UTF-32LE";
#else
static const gchar utf32_byteorder_str[] = "UTF-32BE";
#endif

#define UNICODE_INVALID		 0
#define UNICODE_AMPERSAND	38

typedef	gint (*StrNCmpFunc)(const gchar*, const gchar*, gsize);

typedef struct
{
	guint16		unichr;
	guint16		slen;
	const gchar	*str;

} entity_t;

static const entity_t entities_tab[] =
{
	/*
		-------------------
		ISO 8859-1 entities
		-------------------
	*/

	{160, 4, "nbsp"},	/* no-break space = non-breaking space */
	{161, 5, "iexcl"},	/* inverted exclamation mark */
	{162, 4, "cent"},	/* cent sign */
	{163, 5, "pound"},	/* pound sign */
	{164, 6, "curren"},	/* currency sign */
	{165, 3, "yen"},	/* yen sign = yuan sign */
	{166, 6, "brvbar"},	/* broken bar = broken vertical bar */
	{167, 4, "sect"},	/* section sign */
	{168, 3, "uml"},	/* diaeresis = spacing diaeresis */
	{169, 4, "copy"},	/* copyright sign */
	{170, 4, "ordf"},	/* feminine ordinal indicator */
	{171, 5, "laquo"},	/* left-pointing double angle quotation mark
								= left pointing guillemet */
	{172, 3, "not"},	/* not sign */
	{173, 3, "shy"},	/* soft hyphen = discretionary hyphen */
	{174, 3, "reg"},	/* registered sign = registered trade mark sign */
	{175, 4, "macr"},	/* macron = spacing macron = overline = APL overbar */
	{176, 3, "deg"},	/* degree sign */
	{177, 6, "plusmn"},	/* plus-minus sign = plus-or-minus sign */
	{178, 4, "sup2"},	/* superscript two = superscript digit two = squared */
	{179, 4, "sup3"},	/* superscript three = superscript digit three = cubed */
	{180, 5, "acute"},	/* acute accent = spacing acute */
	{181, 5, "micro"},	/* micro sign */
	{182, 4, "para"},	/* pilcrow sign = paragraph sign */
	{183, 6, "middot"},	/* middle dot = Georgian comma = Greek middle dot */
	{184, 5, "cedil"},	/* cedilla = spacing cedilla */
	{185, 4, "sup1"},	/* superscript one = superscript digit one */
	{186, 4, "ordm"},	/* masculine ordinal indicator */
	{187, 5, "raquo"},	/* right-pointing double angle quotation mark
								= right pointing guillemet */
	{188, 6, "frac14"},	/* vulgar fraction one quarter = fraction one quarter */
	{189, 6, "frac12"},	/* vulgar fraction one half = fraction one half */
	{190, 6, "frac34"},	/* vulgar fraction three quarters = fraction three quarters */
	{191, 6, "iquest"},	/* inverted question mark = turned question mark */
	{192, 6, "Agrave"},	/* latin capital letter A with grave
								= latin capital letter A grave */
	{193, 6, "Aacute"},	/* latin capital letter A with acute */
	{194, 5, "Acirc"},	/* latin capital letter A with circumflex */
	{195, 6, "Atilde"},	/* latin capital letter A with tilde */
	{196, 4, "Auml"},	/* latin capital letter A with diaeresis */
	{197, 5, "Aring"},	/* latin capital letter A with ring above
								= latin capital letter A ring */
	{198, 5, "AElig"},	/* latin capital letter AE = latin capital ligature AE */
	{199, 6, "Ccedil"},	/* latin capital letter C with cedilla */
	{200, 6, "Egrave"},	/* latin capital letter E with grave */
	{201, 6, "Eacute"},	/* latin capital letter E with acute */
	{202, 5, "Ecirc"},	/* latin capital letter E with circumflex */
	{203, 4, "Euml"},	/* latin capital letter E with diaeresis */
	{204, 6, "Igrave"},	/* latin capital letter I with grave */
	{205, 6, "Iacute"},	/* latin capital letter I with acute */
	{206, 5, "Icirc"},	/* latin capital letter I with circumflex */
	{207, 4, "Iuml"},	/* latin capital letter I with diaeresis */
	{208, 3, "ETH"},	/* latin capital letter ETH */
	{209, 6, "Ntilde"},	/* latin capital letter N with tilde */
	{210, 6, "Ograve"},	/* latin capital letter O with grave */
	{211, 6, "Oacute"},	/* latin capital letter O with acute */
	{212, 5, "Ocirc"},	/* latin capital letter O with circumflex */
	{213, 6, "Otilde"},	/* latin capital letter O with tilde */
	{214, 4, "Ouml"},	/* latin capital letter O with diaeresis */
	{215, 5, "times"},	/* multiplication sign */
	{216, 6, "Oslash"},	/* latin capital letter O with stroke
							= latin capital letter O slash */
	{217, 6, "Ugrave"},	/* latin capital letter U with grave */
	{218, 6, "Uacute"},	/* latin capital letter U with acute */
	{219, 5, "Ucirc"},	/* latin capital letter U with circumflex */
	{220, 4, "Uuml"},	/* latin capital letter U with diaeresis */
	{221, 6, "Yacute"},	/* latin capital letter Y with acute */
	{222, 5, "THORN"},	/* latin capital letter THORN */
	{223, 5, "szlig"},	/* latin small letter sharp s = ess-zed */
	{224, 6, "agrave"},	/* latin small letter a with grave = latin small letter a grave */
	{225, 6, "aacute"},	/* latin small letter a with acute */
	{226, 5, "acirc"},	/* latin small letter a with circumflex */
	{227, 6, "atilde"},	/* latin small letter a with tilde */
	{228, 4, "auml"},	/* latin small letter a with diaeresis */
	{229, 5, "aring"},	/* latin small letter a with ring above
								= latin small letter a ring */
	{230, 5, "aelig"},	/* latin small letter ae = latin small ligature ae */
	{231, 6, "ccedil"},	/* latin small letter c with cedilla */
	{232, 6, "egrave"},	/* latin small letter e with grave */
	{233, 6, "eacute"},	/* latin small letter e with acute */
	{234, 5, "ecirc"},	/* latin small letter e with circumflex */
	{235, 4, "euml"},	/* latin small letter e with diaeresis */
	{236, 6, "igrave"},	/* latin small letter i with grave */
	{237, 6, "iacute"},	/* latin small letter i with acute */
	{238, 5, "icirc"},	/* latin small letter i with circumflex */
	{239, 4, "iuml"},	/* latin small letter i with diaeresis */
	{240, 3, "eth"},	/* latin small letter eth */
	{241, 6, "ntilde"},	/* latin small letter n with tilde */
	{242, 6, "ograve"},	/* latin small letter o with grave */
	{243, 6, "oacute"},	/* latin small letter o with acute */
	{244, 5, "ocirc"},	/* latin small letter o with circumflex */
	{245, 6, "otilde"},	/* latin small letter o with tilde */
	{246, 4, "ouml"},	/* latin small letter o with diaeresis */
	{247, 6, "divide"},	/* division sign */
	{248, 6, "oslash"},	/* latin small letter o with stroke */
	{249, 6, "ugrave"},	/* latin small letter u with grave */
	{250, 6, "uacute"},	/* latin small letter u with acute */
	{251, 5, "ucirc"},	/* latin small letter u with circumflex */
	{252, 4, "uuml"},	/* latin small letter u with diaeresis */
	{253, 6, "yacute"},	/* latin small letter y with acute */
	{254, 5, "thorn"},	/* latin small letter thorn with */
	{255, 4, "yuml"},	/* latin small letter y with diaeresis */

	/*
		-----------------------------------------------
		Symbols, mathematical symbols and Greek letters
		-----------------------------------------------
	*/

	{402,  4, "fnof"},	/* latin small f with hook = function */
	{913,  5, "Alpha"},	/* greek capital letter alpha */
	{914,  4, "Beta"},	/* greek capital letter beta */
	{915,  5, "Gamma"},	/* greek capital letter gamma */
	{916,  5, "Delta"},	/* greek capital letter delta */
	{917,  7, "Epsilon"},	/* greek capital letter epsilon */
	{918,  4, "Zeta"},	/* greek capital letter zeta */
	{919,  3, "Eta"},	/* greek capital letter eta */
	{920,  5, "Theta"},	/* greek capital letter theta */
	{921,  4, "Iota"},	/* greek capital letter iota */
	{922,  5, "Kappa"},	/* greek capital letter kappa */
	{923,  6, "Lambda"},	/* greek capital letter lambda */
	{924,  2, "Mu"},	/* greek capital letter mu */
	{925,  2, "Nu"},	/* greek capital letter nu */
	{926,  2, "Xi"},	/* greek capital letter xi */
	{927,  7, "Omicron"},	/* greek capital letter omicron */
	{928,  2, "Pi"},	/* greek capital letter pi */
	{929,  3, "Rho"},	/* greek capital letter rho */
	{931,  5, "Sigma"},	/* greek capital letter sigma */
	{932,  3, "Tau"},	/* greek capital letter tau */
	{933,  7, "Upsilon"},	/* greek capital letter upsilon */
	{934,  3, "Phi"},	/* greek capital letter phi */
	{935,  3, "Chi"},	/* greek capital letter chi */
	{936,  3, "Psi"},	/* greek capital letter psi */
	{937,  5, "Omega"},	/* greek capital letter omega */
	{945,  5, "alpha"},	/* greek small letter alpha */
	{946,  4, "beta"},	/* greek small letter beta */
	{947,  5, "gamma"},	/* greek small letter gamma */
	{948,  5, "delta"},	/* greek small letter delta */
	{949,  7, "epsilon"},	/* greek small letter epsilon */
	{950,  4, "zeta"},	/* greek small letter zeta */
	{951,  3, "eta"},	/* greek small letter eta */
	{952,  5, "theta"},	/* greek small letter theta */
	{953,  4, "iota"},	/* greek small letter iota */
	{954,  5, "kappa"},	/* greek small letter kappa */
	{955,  6, "lambda"},	/* greek small letter lambda */
	{956,  2, "mu"},	/* greek small letter mu */
	{957,  2, "nu"},	/* greek small letter nu */
	{958,  2, "xi"},	/* greek small letter xi */
	{959,  7, "omicron"},	/* greek small letter omicron */
	{960,  2, "pi"},	/* greek small letter pi */
	{961,  3, "rho"},	/* greek small letter rho */
	{962,  6, "sigmaf"},	/* greek small letter final sigma */
	{963,  5, "sigma"},	/* greek small letter sigma */
	{964,  3, "tau"},	/* greek small letter tau */
	{965,  7, "upsilon"},	/* greek small letter upsilon */
	{966,  3, "phi"},	/* greek small letter phi */
	{967,  3, "chi"},	/* greek small letter chi */
	{968,  3, "psi"},	/* greek small letter psi */
	{969,  5, "omega"},	/* greek small letter omega */
	{977,  8, "thetasym"},	/* greek small letter theta symbol */
	{978,  5, "upsih"},	/* greek upsilon with hook symbol */
	{982,  3, "piv"},	/* greek pi symbol */
	{8226, 4, "bull"},	/* bullet = black small circle */
	{8230, 6, "hellip"},	/* horizontal ellipsis = three dot leader */
	{8242, 5, "prime"},	/* prime = minutes = feet */
	{8243, 5, "Prime"},	/* double prime = seconds = inches */
	{8254, 5, "oline"},	/* overline = spacing overscore */
	{8260, 5, "frasl"},	/* fraction slash */
	{8472, 6, "weierp"},	/* script capital P = power set */
	{8465, 5, "image"},	/* blackletter capital I = imaginary part */
	{8476, 4, "real"},	/* blackletter capital R = real part symbol */
	{8482, 5, "trade"},	/* trade mark sign */
	{8501, 8, "alefsym"},	/* alef symbol = first transfinite cardinal */
	{8592, 4, "larr"},	/* leftwards arrow */
	{8593, 4, "uarr"},	/* upwards arrow */
	{8594, 4, "rarr"},	/* rightwards arrow */
	{8595, 4, "darr"},	/* downwards arrow */
	{8596, 4, "harr"},	/* left right arrow */
	{8629, 5, "crarr"},	/* downwards arrow with corner leftwards = carriage return */
	{8656, 4, "lArr"},	/* leftwards double arrow */
	{8657, 4, "uArr"},	/* upwards double arrow */
	{8658, 4, "rArr"},	/* rightwards double arrow */
	{8659, 4, "dArr"},	/* downwards double arrow */
	{8660, 4, "hArr"},	/* left right double arrow */
	{8704, 6, "forall"},	/* for all */
	{8706, 4, "part"},	/* partial differential */
	{8707, 5, "exist"},	/* there exists */
	{8709, 5, "empty"},	/* empty set = null set = diameter */
	{8711, 5, "nabla"},	/* nabla = backward difference */
	{8712, 4, "isin"},	/* element of */
	{8713, 5, "notin"},	/* not an element of */
	{8715, 2, "ni"},	/* contains as member */
	{8719, 4, "prod"},	/* n-ary product = product sign */
	{8721, 3, "sum"},	/* n-ary sumation */
	{8722, 5, "minus"},	/* minus sign */
	{8727, 6, "lowast"},	/* asterisk operator */
	{8730, 5, "radic"},	/* square root = radical sign */
	{8733, 4, "prop"},	/* proportional to */
	{8734, 5, "infin"},	/* infinity */
	{8736, 3, "ang"},	/* angle */
	{8743, 3, "and"},	/* logical and = wedge */
	{8744, 2, "or"},	/* logical or = vee */
	{8745, 3, "cap"},	/* intersection = cap */
	{8746, 3, "cup"},	/* union = cup */
	{8747, 3, "int"},	/* integral */
	{8756, 6, "there4"},	/* therefore */
	{8764, 3, "sim"},	/* tilde operator = varies with = similar to */
	{8773, 4, "cong"},	/* approximately equal to */
	{8776, 5, "asymp"},	/* almost equal to = asymptotic to */
	{8800, 2, "ne"},	/* not equal to */
	{8801, 5, "equiv"},	/* identical to */
	{8804, 2, "le"},	/* less-than or equal to */
	{8805, 2, "ge"},	/* greater-than or equal to */
	{8834, 3, "sub"},	/* subset of */
	{8835, 3, "sup"},	/* superset of */
	{8836, 4, "nsub"},	/* not a subset of */
	{8838, 4, "sube"},	/* subset of or equal to */
	{8839, 4, "supe"},	/* superset of or equal to */
	{8853, 5, "oplus"},	/* circled plus = direct sum */
	{8855, 6, "otimes"},	/* circled times = vector product */
	{8869, 4, "perp"},	/* up tack = orthogonal to = perpendicular */
	{8901, 4, "sdot"},	/* dot operator */
	{8968, 5, "lceil"},	/* left ceiling = apl upstile */
	{8969, 5, "rceil"},	/* right ceiling */
	{8970, 6, "lfloor"},	/* left floor = apl downstile */
	{8971, 6, "rfloor"},	/* right floor */
	{9001, 4, "lang"},	/* left-pointing angle bracket = bra */
	{9002, 4, "rang"},	/* right-pointing angle bracket = ket */
	{9674, 3, "loz"},	/* lozenge */
	{9824, 6, "spades"},	/* black spade suit */
	{9827, 5, "clubs"},	/* black club suit = shamrock */
	{9829, 6, "hearts"},	/* black heart suit = valentine */
	{9830, 5, "diams"},	/* black diamond suit */

	/*
		------------------------------------------------------
		Markup-significant and internationalization characters
		------------------------------------------------------
	*/

	{34,   4, "quot"},	/* quotation mark = APL quote */
	{38,   3, "amp"},	/* ampersand */
	{60,   2, "lt"},	/* less-than sign */
	{62,   2, "gt"},	/* greater-than sign */
	{338,  5, "OElig"},	/* latin capital ligature OE */
	{339,  5, "oelig"},	/* latin small ligature oe */
	{352,  6, "Scaron"},	/* latin capital letter S with caron */
	{353,  6, "scaron"},	/* latin small letter s with caron */
	{376,  4, "Yuml"},	/* latin capital letter Y with diaeresis */
	{710,  4, "circ"},	/* modifier letter circumflex accent */
	{732,  5, "tilde"},	/* small tilde */
	{8194, 4, "ensp"},	/* en space */
	{8195, 4, "emsp"},	/* em space */
	{8201, 6, "thinsp"},	/* thin space */
	{8204, 4, "zwnj"},	/* zero width non-joiner */
	{8205, 3, "zwj"},	/* zero width joiner */
	{8206, 3, "lrm"},	/* left-to-right mark */
	{8207, 3, "rlm"},	/* right-to-left mark */
	{8211, 5, "ndash"},	/* en dash */
	{8212, 5, "mdash"},	/* em dash */
	{8216, 5, "lsquo"},	/* left single quotation mark */
	{8217, 5, "rsquo"},	/* right single quotation mark */
	{8218, 5, "sbquo"},	/* single low-9 quotation mark */
	{8220, 5, "ldquo"},	/* left double quotation mark */
	{8221, 5, "rdquo"},	/* right double quotation mark */
	{8222, 5, "bdquo"},	/* double low-9 quotation mark */
	{8224, 6, "dagger"},	/* dagger */
	{8225, 6, "Dagger"},	/* double dagger */
	{8240, 6, "permil"},	/* per mille sign */
	{8249, 6, "lsaquo"},	/* single left-pointing angle quotation mark */
	{8250, 6, "rsaquo"},	/* single right-pointing angle quotation mark */
	{0, 0, NULL}
};


static UNICODE_CHAR_T cer_to_unichr(const gchar *str, gsize len, gint case_sensitive)
{
	const entity_t	*eptr;
	StrNCmpFunc	str_ncmp_func;

	str_ncmp_func = (StrNCmpFunc)
		(case_sensitive == FALSE) ? GuiStrNCaseCmp : strncmp;

	eptr = entities_tab;

	do
	{
		if (len == eptr->slen &&
			(*str_ncmp_func)(eptr->str, str, len) == 0)
		{
			return eptr->unichr;
		}
	}
	while ((++eptr)->str);

	return UNICODE_INVALID;

} /* static UNICODE_CHAR_T cer_to_unichr(const gchar*, gsize, gint) */

static UNICODE_CHAR_T ncr_to_unichr(const gchar *str, gchar **eptr)
{
	guint 	base, chr;
	gulong	result;

	chr = *(guchar *) str;

	if (isalpha(chr))
	{
		if ((chr != 'x' && chr != 'X') || isxdigit(*(++str)) == FALSE)
		{
			return UNICODE_INVALID;
		}

		base = 16;
	}

	else
	{
		if (isdigit(chr) == FALSE)
		{
			return UNICODE_INVALID;
		}

		base = 10;
	}

	result = strtoul(str, (gchar **) eptr, base);

	if (result && **eptr == ';')
	{
		return result;
	}

	return UNICODE_INVALID;

} /* static UNICODE_CHAR_T ncr_to_unichr(const gchar*, gchar**) */

static gchar *unichr_to_txtstr(UNICODE_CHAR_T unichr, const gchar *charset, gsize *len)
{
	if (unichr == UNICODE_INVALID)
	{
		return NULL;
	}

	*len = sizeof(UNICODE_CHAR_T);

	return charconv((gchar *) &unichr, charset, utf32_byteorder_str, len);
}

static entity_conv_t *txtstr(UNICODE_CHAR_T unichr, const gchar *str, gsize len)
{
	entity_conv_t	*ptr;
	const entity_t	*eptr;

	if (str == NULL)
	{
		eptr = entities_tab;

		do
		{
			if (unichr == eptr->unichr)
			{
				break;
			}
		}
		while ((++eptr)->str);

		if ((str = eptr->str) == NULL)
		{
			return NULL;
		}

		len = eptr->slen;
	}

	len++;

	ptr = g_malloc(sizeof(entity_conv_t) + len);
	ptr->blen = len;
	memcpy(ptr->bstr, str, len);
	ptr->bstr[len - 1] = ';';
	ptr->bstr[len] = '\0';

	return ptr;

} /* static entity_conv_t *txtstr(UNICODE_CHAR_T, const gchar*, gsize) */

static entity_conv_t *numstr(UNICODE_CHAR_T unichr)
{
	gint		len;
	gchar		buf[64];
	entity_conv_t	*ptr;

	len = g_snprintf(buf, sizeof(buf) - 1, "#%lu;", (gulong) unichr);

	if (len < 0)
	{
		return NULL;
	}

	ptr = g_malloc(sizeof(entity_conv_t) + len);
	ptr->blen = len;
	ptr->bstr[len] = '\0';
	memcpy(ptr->bstr, buf, len);

	return ptr;

} /* static entity_conv_t *numstr(UNICODE_CHAR_T) */

void xsearch_get_entities_doc(xsearch_t *xptr)
{
	gsize		count, pos, mlen, ulen;
	UNICODE_CHAR_T	*unistr_lo, *unistr_up, *unistr[2];
	entity_conv_t	**dtab;

	xptr->entities_doc = NULL;

	ulen = xptr->plen;
	unistr_lo = (UNICODE_CHAR_T *)
			charconv(xptr->str_lo, utf32_byteorder_str, xptr->charset, &ulen);

	if (unistr_lo == NULL)
	{
		return;
	}

	if (xptr->str_lo == xptr->str_up)
	{
		unistr_up = unistr_lo;
	}

	else if (xptr->str_lo == (gchar *) unistr_lo)
	{
		unistr_up = (UNICODE_CHAR_T *) xptr->str_up;
	}

	else
	{
		ulen = xptr->plen;
		unistr_up = (UNICODE_CHAR_T *)
			charconv(xptr->str_up, utf32_byteorder_str, xptr->charset, &ulen);

		if (unistr_up == NULL)
		{
			if (xptr->str_lo != (gchar *) unistr_lo)
			{
				g_free(unistr_lo);
			}

			return;
		}
	}

	unistr[0] = unistr_lo;
	unistr[1] = unistr_up;

	count	= 0;
	pos	= 0;
	mlen	= ulen / sizeof(UNICODE_CHAR_T);

	ulen = mlen << 1;

	if (xptr->str_lo != xptr->str_up)
	{
		ulen <<= 1;
	}

	dtab = g_malloc((ulen + 1) * sizeof(gpointer));

	do
	{
		gint		n, loop;
		UNICODE_CHAR_T	unichr_lo, unichr_up;
		const entity_t	*eptr;

		unichr_lo = *(unistr_lo++);
		unichr_up = *(unistr_up++);

		if (unichr_lo == UNICODE_AMPERSAND)
		{
			continue;
		}

		eptr	= entities_tab;
		n	= 0;
		loop	= 2;

		do
		{
			gchar		*ptr, *str;
			gsize		len;
			UNICODE_CHAR_T	unichr;

			if (unichr_lo == eptr->unichr)
			{
				unichr = unichr_lo;
				str = xptr->str_lo;
			}

			else if (unichr_up == eptr->unichr)
			{
				unichr = unichr_up;
				str = xptr->str_up;
			}

			else
			{
				continue;
			}

			ptr = unichr_to_txtstr(unichr, xptr->charset, &len);

			if (ptr)
			{
				gchar *s = tl_mem_find_str(&str[pos],
							xptr->plen - pos, ptr, len);

				if (ptr != (gchar *) unichr)
				{
					g_free(ptr);
				}

				if (s)
				{
					gint i = 2;

					do
					{
						entity_conv_t *dptr = (i == 2) ?
							txtstr(unichr, eptr->str, eptr->slen) :
										numstr(unichr);
						if (dptr)
						{
							dtab[count] = dptr;
							count++;

							n = s - str;
							dptr->pos = n;
							dptr->plen = len;
						}
					}
					while (--i > 0);
				}

				if (unichr_lo == unichr_up || --loop == 0)
				{
					if (n > 0)
					{
						pos = n + 1;
					}

					break;
				}
			}
		}
		while ((++eptr)->str);
	}
	while (--mlen > 0);

	if (xptr->str_lo != (gchar *) unistr[0])
	{
		g_free(unistr[0]);

		if (xptr->str_lo != xptr->str_up)
		{
			g_free(unistr[1]);
		}
	}

	if(count > 0)
	{
		dtab[count] = NULL;
		xptr->entities_doc = g_realloc(dtab, (count + 1) * sizeof(gpointer));
	}

	else
	{
		g_free(dtab);
	}

} /* void xsearch_get_entities_doc(xsearch_t*) */

void xsearch_get_entities_pat(xsearch_t *xptr)
{
	gint		case_sensitive;
	guint		ncrchr;
	gsize		count_chr, count_str, sz;
	gchar		*end, *pattern;
	entity_conv_t	**ptab_chr, **ptab_str;

	xptr->entities_pat_chr = NULL;
	xptr->entities_pat_str = NULL;

	case_sensitive = (xptr->str_lo == xptr->str_up);

	sz = xptr->plen;

	if (case_sensitive == FALSE)
	{
		sz <<= 1;
	}

	ptab_chr = g_malloc((sz + 1) * sizeof(gpointer));
	ptab_str = g_malloc(((sz << 1) + 1) * sizeof(gpointer));

	count_chr = 0;
	count_str = 0;
	ncrchr    = 0;

	pattern	= xptr->str_lo;
	end	= pattern;

	while ((pattern = memchr(end, '&', &xptr->str_lo[xptr->plen] - end)) != NULL)
	{
		gint		loop, type;
		gsize		plen;
		UNICODE_CHAR_T	unichr, unichr_orig, unichr_lo, unichr_up;

		end = memchr(&pattern[1], ';', &xptr->str_lo[xptr->plen] - &pattern[1]);

		if (end == NULL)
		{
			break;
		}

		plen = end - pattern;

		unichr_orig = (pattern[1] == '#') ?
			ncr_to_unichr(&pattern[2], &end) :
				cer_to_unichr(&pattern[1], plen - 1, case_sensitive);

		if (unichr_orig == UNICODE_INVALID)
		{
			continue;
		}

		loop = 3;

		if (case_sensitive == FALSE)
		{
			unichr_lo = UNICODE_TOOLS_TOLOWER(unichr_orig);
			unichr_up = UNICODE_TOOLS_TOUPPER(unichr_orig);

			if (unichr_lo != unichr_up)
			{
				loop = 6;
			}
		}

		else
		{
			unichr_lo = unichr_orig;
			unichr_up = unichr_orig;
		}

		type   = 3;
		unichr = unichr_lo;

		do
		{
			entity_conv_t *pptr;

			if (loop == 3 && unichr_lo != unichr_up)
			{
				type   = 3;
				unichr = unichr_up;
			}

			if (type == 3)
			{
				if (pattern[1] != '#')
				{
					continue;
				}

				pptr = txtstr(unichr, NULL, 0);
			}

			else if (type == 2)
			{
				if (pattern[1] == '#')
				{
					ncrchr = '#';

					if (unichr == unichr_orig &&
						pattern[2] != 'x' && pattern[2] != 'X')
					{
						continue;
					}
				}

				pptr = numstr(unichr);
			}

			else
			{
				gsize  slen;
				gchar  *str;

				if (unichr == UNICODE_AMPERSAND ||
					UNICODE_TOOLS_ISPRINT(unichr) == FALSE)
				{
					continue;
				}

				str = unichr_to_txtstr(unichr, xptr->charset, &slen);

				if (str == NULL)
				{
					continue;
				}

				pptr = g_malloc(sizeof(entity_conv_t) + slen);

				pptr->blen = slen;
				pptr->bstr[slen] = '\0';
				memcpy(pptr->bstr, str, slen);

				if (str != (gchar *) unichr)
				{
					g_free(str);
				}
			}

			if (pptr)
			{
				pptr->pos  = pattern - xptr->str_lo;
				pptr->plen = plen + 1;

				if (type > 1)
				{
					/* skip '&' */
					pptr->pos++;
					pptr->plen--;

					ptab_str[count_str] = pptr;
					count_str++;
				}

				else
				{
					ptab_chr[count_chr] = pptr;
					count_chr++;
				}
			}
		}
		while (--type, --loop > 0);

	} /* while ((pattern = (memchr(pattern, '&', &xptr->str_lo[xptr->plen] - end) != NULL) */

	if (count_chr > 0)
	{
		ptab_chr[count_chr] = NULL;
		xptr->entities_pat_chr = g_realloc(ptab_chr,
						(count_chr + 1) * sizeof(gpointer));
	}

	else
	{
		g_free(ptab_chr);
	}

	if (count_str > 0 || ncrchr == '#')
	{
		ptab_str[count_str] = NULL;
		xptr->entities_pat_str = g_realloc(ptab_str,
						(count_str + 1) * sizeof(gpointer));
	}

	else
	{
		g_free(ptab_str);
	}

} /* void xsearch_get_entities_pat(xsearch_t*) */

void xsearch_free_entities(xsearch_t *xptr)
{
	entity_conv_t **entities, *entity;

	if ((entities = xptr->entities_doc) != NULL)
	{
		while ((entity = *(entities++)) != NULL)
		{
			g_free(entity);
		}

		g_free(xptr->entities_doc);
	}

	if ((entities = xptr->entities_pat_chr) != NULL)
	{
		while ((entity = *(entities++)) != NULL)
		{
				g_free(entity);
		}

		g_free(xptr->entities_pat_chr);
	}

	if ((entities = xptr->entities_pat_str) != NULL)
	{
		while ((entity = *(entities++)) != NULL)
		{
				g_free(entity);
		}

		g_free(xptr->entities_pat_str);
	}

} /* void xsearch_free_entities(xsearch_t*) */

gchar *xsearch_mem_find_str(const xsearch_t *xptr, const gchar *mem, gint mlen, gint *end_offs)
{
	guint		c, c_lo, c_up, pos;
	const guchar	*beg, *stop, *str_lo, *str_up;

	if (mem == NULL || mlen <= 0)
	{
		return NULL;
	}

	stop = &mem[mlen];

	str_lo = xptr->str_lo;
	str_up = xptr->str_up;
	c_lo = *str_lo;
	c_up = *str_up;
	pos = 0;
	beg = mem;

#ifdef PRE_INCREMENT

	beg--;

	while (c = *(++beg), stop > beg)
#else
	while (c = *(beg++), stop >= beg)
#endif
	{
		gsize		max;
		const guchar	*s;
		entity_conv_t	**ptab;

		if (c == c_lo || c == c_up)
		{
			if (pos == 0)
			{
				mem = beg;
			}

			pos++;
__match__:
			if (pos == xptr->plen)
			{
#ifdef PRE_INCREMENT
				beg++;
#else
				mem--;
#endif
				*end_offs = beg - (guchar *) mem;

				return (gchar *) mem;
			}

			c_lo = str_lo[pos];
			c_up = str_up[pos];
			continue;
		}

		else if (c == '&')
		{
			entity_conv_t **dtab;
#ifdef PRE_INCREMENT
			s = &beg[1];
#else
			s = &beg[0];
#endif
			if ((dtab = xptr->entities_doc) != NULL)
			{
				entity_conv_t *dptr;

				max = stop - s;

				while ((dptr = *(dtab++)) != NULL && pos >= dptr->pos)
				{
					if (pos == dptr->pos &&
						max >= dptr->blen &&
							memcmp(s, dptr->bstr,
								dptr->blen) == 0)
					{
						if (pos == 0)
						{
							mem = beg;
						}

						beg += dptr->blen;
						pos += dptr->plen;

						goto __match__;
					}
				}

			} /* if ((dtab = xptr->entities_doc) != NULL) */

			if (*s == '#')
			{
				gint 		len;
				gchar		*str, *eptr;
				UNICODE_CHAR_T	unichr;

				unichr = ncr_to_unichr(&s[1], &eptr);
				str = unichr_to_txtstr(unichr, xptr->charset, &len);

				if (str)
				{
					gint match = FALSE;

					if (len == 1)
					{
						c = *(guchar *) str;

						if (c == c_lo || c == c_up)
						{
							match = TRUE;
						}
					}

					else if ( len + pos <= xptr->plen &&
						(memcmp(&str_lo[pos], str, len) == 0 ||
							memcmp(&str_up[pos], str, len) == 0) )
					{
						match = TRUE;
					}

					if (str != (gchar *) &unichr)
					{
						g_free(str);
					}

					if (match)
					{
						if (pos == 0)
						{
							mem = beg;
						}
#ifdef PRE_INCREMENT
						beg = eptr;
#else
						beg = &eptr[1];
#endif
						pos += len;

						goto __match__;
					}

				} /* if (s) */

			} /* if (*s == '#') */

		} /* else if (c == '&') */

		else
		{
			entity_conv_t *pptr;

			if ((ptab = xptr->entities_pat_chr) != NULL)
			{
#ifdef PRE_INCREMENT
				s = &beg[0];
#else
				s = &beg[-1];
#endif
				max = stop - s;

				while ((pptr = *(ptab++)) != NULL && pos >= pptr->pos)
				{
					if (pos == pptr->pos)
					{
						if (pptr->blen == 1)
						{
							if (c == *pptr->bstr)
							{
								goto __pat_match__;
							}
						}

						else if (max >= pptr->blen &&
							memcmp(s, pptr->bstr, pptr->blen) == 0)
						{
__pat_match__:
								if (pos == 0)
								{
									mem = beg;
								}

								beg += pptr->blen - 1;
								pos += pptr->plen;

								goto __match__;
						}
					}
				}

			} /* if ((ptab = xptr->entities_pat_chr) != NULL) */

			if (pos > 0 && (ptab = xptr->entities_pat_str) != NULL)
			{
#ifdef PRE_INCREMENT
				s = &beg[0];
#else
				s = &beg[-1];
#endif
				if (s[-1] != '&')
				{
					c = *s;

					do
					{
						if (--pos == 0 ||
							(isxdigit(c) == FALSE &&
								c != 'x' && c != 'X'))
						{
							goto __no_match__;
						}
					}
					while ((c = *(--s)) != '#');

					if (s[-1] != '&')
					{
						goto __no_match__;
					}
				}

				max = stop - s;

				while ((pptr = *(ptab++)) != NULL && pos >= pptr->pos)
				{
					if (pos == pptr->pos && max >= pptr->blen &&
							memcmp(s, pptr->bstr, pptr->blen) == 0)
					{
#ifdef PRE_INCREMENT
						beg = &s[pptr->blen - 1];
#else
						beg = &s[pptr->blen];
#endif
						pos += pptr->plen;

						goto __match__;
					}
				}

				if (*s == '#')
				{
					gchar		*eptr_doc, *eptr_pat;
					UNICODE_CHAR_T	unichr_doc, unichr_pat;

					unichr_doc = ncr_to_unichr(&s[1], &eptr_doc);

					if (unichr_doc != UNICODE_INVALID)
					{
						gint  case_sensitive;
						gchar *str;

						case_sensitive = (str_lo == str_up);
						str = (gchar *) &str_lo[pos];

						if (*str == '#')
						{
							str++;
							unichr_pat =
							ncr_to_unichr(str, &eptr_pat);
						}

						else
						{
							eptr_pat = memchr(str, ';',
									xptr->plen - pos);
							unichr_pat = cer_to_unichr(str,
								(eptr_pat - str), case_sensitive);
						}

						if (unichr_pat != UNICODE_INVALID)
						{
							if (case_sensitive == FALSE)
							{
								unichr_doc =
									UNICODE_TOOLS_TOLOWER(unichr_doc);
								unichr_pat =
									UNICODE_TOOLS_TOLOWER(unichr_pat);
							}

							if (unichr_doc == unichr_pat)
							{
#ifdef PRE_INCREMENT
								beg = eptr_doc;
#else
								beg = &eptr_doc[1];
#endif
								pos = ++eptr_pat - (gchar *) str_lo;

								goto __match__;
							}
						
						} /* if (unichr_pat != UNICODE_INVALID) */

					} /* if (unichr_doc) */

				} /* if (*s == '#') */

			} /* if (pos > 0 && (ptab = xptr->entities_pat_str) != NULL)) */

		} /* else if (c == '&') .. else */

		if (pos > 0)
		{
__no_match__:
			pos  = 0;
			beg  = mem;
			c_lo = *str_lo;
			c_up = *str_up;
		}

	} /* while (c = *(beg++), stop >= beg) */

	return NULL;

} /* gchar *xsearch_mem_find_str(const xsearch_t*, const gchar*, gint, gint*) */

#endif /* HAVE_XSEARCH */

/* EOF */
